#include "scrabble.h"
#include "dict.h"
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#if 0
/* english letters */
letter letters[] = {
  {   0, 0, 2 },           /* joker */
  { 'A', 1, 9 },
  { 'B', 3, 2 },
  { 'C', 3, 2 },
  { 'D', 2, 4 },
  { 'E', 1, 12 },
  { 'F', 4, 2 },
  { 'G', 2, 3 },
  { 'H', 4, 2 },
  { 'I', 1, 9 },
  { 'J', 8, 1 },
  { 'K', 5, 1 },
  { 'L', 1, 4 },
  { 'M', 3, 2 },
  { 'N', 1, 6 },
  { 'O', 1, 8 },
  { 'P', 3, 2 },
  { 'Q', 10, 1 },
  { 'R', 1, 6 },
  { 'S', 1, 4 },
  { 'T', 1, 6 },
  { 'U', 1, 4 },
  { 'V', 4, 2 },
  { 'W', 4, 2 },
  { 'X', 8, 1 },
  { 'Y', 4, 2 },
  { 'Z', 10, 1 },
};
#endif

/* french letters */
letter letters[] = {
  {   0, 0, 2 },           /* joker */
  { 'A', 1, 9 },
  { 'B', 3, 2 },
  { 'C', 3, 2 },
  { 'D', 2, 3 },
  { 'E', 1, 15 },
  { 'F', 4, 2 },
  { 'G', 2, 2 },
  { 'H', 4, 2 },
  { 'I', 1, 8 },
  { 'J', 8, 1 },
  { 'K', 10, 1 },
  { 'L', 1, 5 },
  { 'M', 2, 3 },
  { 'N', 1, 6 },
  { 'O', 1, 6 },
  { 'P', 3, 2 },
  { 'Q', 8, 1 },
  { 'R', 1, 6 },
  { 'S', 1, 6 },
  { 'T', 1, 6 },
  { 'U', 1, 6 },
  { 'V', 4, 2 },
  { 'W', 10, 1 },
  { 'X', 10, 1 },
  { 'Y', 10, 1 },
  { 'Z', 10, 1 },
};

int value[15][15] = {
{ TW,  0,  0, DL,  0,  0,  0, TW,  0,  0,  0, DL,  0,  0, TW  },
{  0, DW,  0,  0,  0, TL,  0,  0,  0, TL,  0,  0,  0, DW,  0  },
{  0,  0, DW,  0,  0,  0, DL,  0, DL,  0,  0,  0, DW,  0,  0  },
{ DL,  0,  0, DW,  0,  0,  0, DL,  0,  0,  0, DW,  0,  0, DL  },
{  0,  0,  0,  0, DW,  0,  0,  0,  0,  0, DW,  0,  0,  0,  0  },
{  0, TL,  0,  0,  0, TL,  0,  0,  0, TL,  0,  0,  0, TL,  0  },
{  0,  0, DL,  0,  0,  0, DL,  0, DL,  0,  0,  0, DL,  0,  0  },
{ TW,  0,  0, DL,  0,  0,  0, DW,  0,  0,  0, DL,  0,  0, TW  },
{  0,  0, DL,  0,  0,  0, DL,  0, DL,  0,  0,  0, DL,  0,  0  },
{  0, TL,  0,  0,  0, TL,  0,  0,  0, TL,  0,  0,  0, TL,  0  },
{  0,  0,  0,  0, DW,  0,  0,  0,  0,  0, DW,  0,  0,  0,  0  },
{ DL,  0,  0, DW,  0,  0,  0, DL,  0,  0,  0, DW,  0,  0, DL  },
{  0,  0, DW,  0,  0,  0, DL,  0, DL,  0,  0,  0, DW,  0,  0  },
{  0, DW,  0,  0,  0, TL,  0,  0,  0, TL,  0,  0,  0, DW,  0  },
{ TW,  0,  0, DL,  0,  0,  0, TW,  0,  0,  0, DL,  0,  0, TW  }};

/* set to 1 for a letter in the board that is a joker */
int is_joker[15][15];

int board[15][15];
int hand[7];
dictionary dict;
dictionary found;

void read_dict(char *fn)
{
  FILE *f = fopen(fn, "r"); if (!f) { perror(fn); abort(); }
  while (1) {
    char n[128];
    int ni[128];
    int i;
    int len;
    if (!fgets(n, 128, f)) break;
    if ((len=strlen(n)) == 0) continue;
    if (n[len-1] == '\n') n[len-1] = 0;
    for (i = 0; n[i]; i++) ni[i] = toupper(n[i])-'A'+1;
    ni[i] = 0;
    add_word(&dict, ni);
  }
  fclose(f);
  return;
}

void read_board(char *fn)
{
  FILE *f = fopen(fn, "r"); if (!f) { perror(fn); abort(); }
  int i, j;
  int c;
  for (i = 0; i < 15; i++)
  for (j = 0; j < 15; j++) {
    do {
      c = fgetc(f); if (c == EOF) goto bf;
    } while (isspace(c));
    if (c == '.') c = 0; else {
      if (!isalpha(c)) goto bf;
      /* lower letters are joker letters (no point) */
      if (islower(c)) is_joker[i][j] = 1;
      c = toupper(c);
    }
    board[i][j] = c ? c - 'A' + 1 : -1000;
  }
  fclose(f);
  return;
bf:
  printf("%s: bad file\n", fn); abort();
}

void read_hand(char *h)
{
  int i;
  if (strlen(h) != 7) goto bh;
  for (i = 0; i < 7; i++) {
    int c = h[i];
    if (c == '_') c = 0;                 /* joker */
    else if (c == '.') c = 0x10000;      /* no letter, mark it as used */
    else { if (!isalpha(c)) goto bh; c = toupper(c) - 'A' + 1; }
    hand[i] = c;
  }
  return;
bh:
  printf("%s: bad hand\n", h); abort();
}

typedef struct {
  dict_letter *l;
  int hl;         /* hand's value (to check for jokers) */
} stack_elem;

int starti, startj, inci, incj;
stack_elem stack[15];
int stacklen;

#if 0
char *getword(void)
{
  int i, j, l;
  static char v[16];
  for (l=0, i=starti, j=startj; l < stacklen; l++, i+=inci, j+=incj)
    v[l] = stack[l].l->l + 'A' - 1 + (board[i][j]!=-1000 ? 'a'-'A' : 0);
  v[l] = 0;
  return v;
}
#endif

int is_letter(int i, int j)
{
  return i >= 0 && i < 15 && j >= 0 && j < 15 && board[i][j] != -1000;
}

int line_score(int i, int j, int inci, int incj)
{
  int mulfac = 1;
  int score = 0;
  int nletters = 0;

  while (is_letter(i-inci, j-incj)) i-=inci, j-=incj;

  while (is_letter(i, j)) {
    int letter_mul = 1;
    if (board[i][j] <= 0) switch (value[i][j]) {
    case DW: mulfac *= 2; break;
    case TW: mulfac *= 3; break;
    case DL: letter_mul = 2; break;
    case TL: letter_mul = 3; break;
    }
    if (!is_joker[i][j])
      score += letters[abs(board[i][j])].value * letter_mul;
    nletters++;
    i+=inci;
    j+=incj;
  }

  if (nletters == 1) return 0;
  return score * mulfac;
}

int score(void)
{
  int s;
  int i, j, l;
  for (l=0, i=starti, j=startj; l<stacklen; l++, i+=inci, j+=incj)
    if (board[i][j] == -1000) board[i][j] = -stack[l].hl;
  s = line_score(starti, startj, inci, incj);
  for (l=0, i=starti, j=startj; l<stacklen; l++, i+=inci, j+=incj) {
    if (board[i][j] > 0) continue;
    s += line_score(i, j, incj, inci);  /* other direction */
  }
  for (l=0, i=starti, j=startj; l<stacklen; l++, i+=inci, j+=incj)
    if (board[i][j] <= 0) board[i][j] = -1000;
  /* if all letters of (full) hand used, +50 */
  i = 0;
  for (l = 0; l < 7; l++) if (hand[l] & 0x10000) i++;
  if (i == 7) s += 50;
  return s;
}

void check_word(void)
{
  /* to be ok a word must include an already positioned letter or touch one
   * or be put on the center position (can only happen for first word)
   * note that when check_word is called we already checked that putting
   * the letters on the board give existing word (or words)
   */
  int w[4+16];
  int i, j, l, ok = 0;
  for (l=0, i=starti, j=startj; l<stacklen; l++, i+=inci, j+=incj)
    if (is_letter(i, j) ||             /* word contain letter on board */
        is_letter(i-incj, j-inci) ||   /* word touch letter on board */
        is_letter(i+incj, j+inci) ||   /* word touch letter on board */
        (i == 7 && j == 7))            /* word is at center pos (first word) */
    { ok = 1; break; }
  if (!ok) return;
  /* we put score, i, j, inci, at start of word, but! +0x1000000 to avoid 0 */
  w[0] = score() + 0x1000000;
  w[1] = starti + 0x1000000;
  w[2] = startj + 0x1000000;
  w[3] = inci + 0x1000000;
  /* letters coming from joker stored with negative value (see print_word) */
  for (l=0; l<stacklen; l++) w[l+4] = stack[l].l->l * (stack[l].hl ? 1 : -1);
  w[l+4] = 0;
  add_word(&found, w);
//  printf("%s (i %d j %d %s) %d\n", getword(), starti, startj, incj?"horiz":"vert", score());
}

void push(dict_letter *n, int hl, int ni, int nj)
{
  stack[stacklen].l = n;
  stack[stacklen].hl = hl;
  stacklen++;
  /* if n->w we are in an end letter of a word in the dictionary
   * so if on the board there is nothing after current letter,
   * we check that the word is ok here at (i,j)
   */
  if (n->w && (ni==15 || nj==15 || board[ni][nj]==-1000)) check_word();
}

void pop(void) { stacklen--; }

int other_word_ok(dict_letter *n, int i, int j)
{
  int si = i, sj = j;
  int word[16];
  int w;
  int ii = 1 - inci;
  int jj = 1 - incj;
  int starti = i, startj = j;
  int endi, endj;
  board[si][sj] = n->l;
  /* find start/end pos of word */
  while (starti >= 0 && startj >= 0 && board[starti][startj] != -1000)
    starti -= ii, startj -= jj;
  starti += ii; startj += jj;
  endi = starti; endj = startj;
  while (endi < 15 && endj < 15 && board[endi][endj] != -1000)
    endi += ii, endj += jj;
  endi -= ii; endj -= jj;
  /* get word */
  for (w=0, i=starti, j=startj; i!=endi+ii || j!=endj+jj; w++, i+=ii, j+=jj)
    word[w] = board[i][j];
  word[w] = 0;
  board[si][sj] = -1000;
//if (w !=1) printf("w %d\n", w);
  return w == 1 || find_word(&dict, word);
}

void go(dict_letter *cur, int i, int j)
{
  /* this function works vertically or horizontally depending on inci/incj */
  /* if in the dictionary the letter at (i,j) matches one in cur,
   * the go to next pos on board and go down in the dictionary tree.
   * the word is checked in push
   */
  int h;

  if (cur == NULL) return;
  if (i == 15 || j == 15) return;

  if (board[i][j] != -1000) {
    /* must match board[i][j] */
    while (cur) {
      if (cur->l == board[i][j]) {
        push(cur, cur->l, i + inci, j + incj);
        go(cur->child, i + inci, j + incj);
        pop();
        return;
      }
      cur = cur->next;
    }
    return;
  }

  /* must match one of not-yet-used letters */
  for (h = 0; h < 7; h++) {
    dict_letter *n = cur;
    int g;
    if (hand[h] & 0x10000) continue;  /* letter already used */
    g = hand[h];
    hand[h] |= 0x10000;
    while (n) {
      /* letter must match or g is joker
       * and (eventual) word in other direction must be ok */
      if ((n->l == g || g == 0) && other_word_ok(n, i, j)) {
        push(n, g, i + inci, j + incj);
        go(n->child, i + inci, j + incj);
        pop();
      }
      n = n->next;
    }
    hand[h] &= ~0x10000;
  }
}

void print_word(int *w)
{
  int i, a, b, inca, incb;
  /* print word */
  for (i = 4; w[i]; i++) printf("%c", abs(w[i]) + 'A' - 1);
  printf(" lin %d col %d %s ",
         w[1] - 0x1000000, w[2] - 0x1000000,
         w[3] - 0x1000000 ? "verti" : "horiz");
  a = w[1] - 0x1000000;
  b = w[2] - 0x1000000;
  inca = w[3] - 0x1000000;
  incb = 1 - inca;
  /* print - for joker letter, UPPER for letter on board (ie. not in hand) */
  for (i = 4; w[i]; i++, a+=inca, b+=incb)
    printf("%c", w[i]<0? '-' : board[a][b] == -1000? w[i]+'a'-1 : w[i]+'A'-1);
  /* score at field 9 so you can sort by score (is -k gnu sort specific?):
   * ./scrabble <dict> <game> <hand>|sort -k 9nr|less
   */
  printf(" score %d\n",
         w[0] - 0x1000000);
}

int main(int n, char **v)
{
#if 1
  int i, j;
  if (n != 4) { printf("gimme <dictionary> <board file> <hand>\n"); return 1; }
  new_dict(&dict);
  new_dict(&found);
  read_dict(v[1]);
  read_board(v[2]);
  read_hand(v[3]);
  inci = 0; incj = 1;
  for (i = 0; i < 15; i++) for (j = 0; j < 15; j++) {
    starti = i; startj = j;
    /* must be a start of a word (no letter before) */
    if (is_letter(starti-inci, startj-incj)) continue;
    go(dict.top, i, j);
  }
  inci = 1; incj = 0;
  for (i = 0; i < 15; i++) for (j = 0; j < 15; j++) {
    starti = i; startj = j;
    /* must be a start of a word (no letter before) */
    if (is_letter(starti-inci, startj-incj)) continue;
    go(dict.top, i, j);
  }
  dump_cb(&found, print_word);
  return 0;
#endif
#if 0
  /* print count of letters (debug stuff) */
  int i;
  n = 0;
  for (i = 0; i < 27; i++) n += letters[i].count;
  printf("%d\n", n); return 0;
#endif
#if 0
  /* output board image (debug stuff for double/triple letter/word)
   * (in array 'value') image format PPM
   */
int col[5][3] = {
  { 20, 110, 20 },
  { 127, 127, 245 },
  { 40, 40, 245 },
  { 215, 127, 127 },
  { 215, 10, 10 }
};

#define WIDTH 20
#define HEIGHT 20
#define BG "30 60 30\n"
  int i, j, I, J;
  printf("P3\n%d %d\n255\n", 15*WIDTH+2, 15*HEIGHT+2);
  printf(BG);
  for (j = 0; j < 15; j++)
  for (J = 0; J < WIDTH; J++)
    printf(BG);
  printf(BG);
  for (i = 0; i < 15; i++)
  for (I = 0; I < HEIGHT; I++) {
    printf(BG);
    for (j = 0; j < 15; j++)
    for (J = 0; J < WIDTH; J++)
      if (I % HEIGHT == 0 || I % HEIGHT == HEIGHT-1 ||
          J % WIDTH == 0 || J % WIDTH == WIDTH-1)
        printf(BG);
      else
        printf("%d %d %d\n", col[value[i][j]][0], col[value[i][j]][1],
                             col[value[i][j]][2]);
    printf(BG);
  }
  printf(BG);
  for (j = 0; j < 15; j++)
  for (J = 0; J < WIDTH; J++)
    printf(BG);
  printf(BG);
  return 0;
#endif
}
